"""Defines the structure for Genie API commands."""
CLEAR_ALL_MEMORY = {"command": "clearAllMemory", "strings": [], "vectors": [], "emotives": {}}
CLEAR_WM = {"command": "clearWM", "strings": [], "vectors": [], "emotives": {}}
LEARN = {"command": "learn", "strings": [], "vectors": [], "emotives": {}}
SET_PREDICTIONS_ON = {"command": "setPredictionsOn", "strings": [], "vectors": [], "emotives": {}}
SET_PREDICTIONS_OFF = {"command": "setPredictionsOff", "strings": [], "vectors": [], "emotives": {}}
START_ACTING = {"command": "startActing", "strings": [], "vectors": [], "emotives": {}}
STOP_ACTING = {"command": "stopActing", "strings": [], "vectors": [], "emotives": {}}
START_SLEEPING = {"command": "startSleeping", "strings": [], "vectors": [], "emotives": {}}
STOP_SLEEPING = {"command": "stopSleeping", "strings": [], "vectors": [], "emotives": {}}
